@@ -2,165 +2,48 @@
 
   objspace.c - ObjectSpace extender for MRI.
 
-  $Author: zzak $
+  $Author$
   created at: Wed Jun 17 07:39:17 2009
 
-  NOTE: This extension library is not expected to exist except C Ruby.
+  NOTE: This extension library is only expected to exist with C Ruby.
 
   All the files in this distribution are covered under the Ruby's
   license (see the file COPYING).
 
 **********************************************************************/
 
-#include <ruby/ruby.h>
-#include <ruby/st.h>
-#include <ruby/io.h>
-#include <ruby/re.h>
-#include "node.h"
 #include "gc.h"
-#include "regint.h"
 #include "internal.h"
-
-size_t rb_str_memsize(VALUE);
-size_t rb_ary_memsize(VALUE);
-size_t rb_io_memsize(const rb_io_t *);
-size_t rb_generic_ivar_memsize(VALUE);
-size_t rb_objspace_data_type_memsize(VALUE obj);
-
-static size_t
-memsize_of(VALUE obj)
-{
-    size_t size = 0;
-
-    if (SPECIAL_CONST_P(obj)) {
-	return 0;
-    }
-
-    if (FL_TEST(obj, FL_EXIVAR)) {
-	size += rb_generic_ivar_memsize(obj);
-    }
-
-    switch (BUILTIN_TYPE(obj)) {
-      case T_OBJECT:
-	if (!(RBASIC(obj)->flags & ROBJECT_EMBED) &&
-	    ROBJECT(obj)->as.heap.ivptr) {
-	    size += ROBJECT(obj)->as.heap.numiv * sizeof(VALUE);
-	}
-	break;
-      case T_MODULE:
-      case T_CLASS:
-	if (RCLASS_M_TBL(obj)) {
-	    size += st_memsize(RCLASS_M_TBL(obj));
-	}
-	if (RCLASS_IV_TBL(obj)) {
-	    size += st_memsize(RCLASS_IV_TBL(obj));
-	}
-	if (RCLASS_IV_INDEX_TBL(obj)) {
-	    size += st_memsize(RCLASS_IV_INDEX_TBL(obj));
-	}
-	if (RCLASS(obj)->ptr->iv_tbl) {
-	    size += st_memsize(RCLASS(obj)->ptr->iv_tbl);
-	}
-	if (RCLASS(obj)->ptr->const_tbl) {
-	    size += st_memsize(RCLASS(obj)->ptr->const_tbl);
-	}
-	size += sizeof(rb_classext_t);
-	break;
-      case T_STRING:
-	size += rb_str_memsize(obj);
-	break;
-      case T_ARRAY:
-	size += rb_ary_memsize(obj);
-	break;
-      case T_HASH:
-	if (RHASH(obj)->ntbl) {
-	    size += st_memsize(RHASH(obj)->ntbl);
-	}
-	break;
-      case T_REGEXP:
-	if (RREGEXP(obj)->ptr) {
-	    size += onig_memsize(RREGEXP(obj)->ptr);
-	}
-	break;
-      case T_DATA:
-	size += rb_objspace_data_type_memsize(obj);
-	break;
-      case T_MATCH:
-	if (RMATCH(obj)->rmatch) {
-            struct rmatch *rm = RMATCH(obj)->rmatch;
-	    size += onig_region_memsize(&rm->regs);
-	    size += sizeof(struct rmatch_offset) * rm->char_offset_num_allocated;
-	    size += sizeof(struct rmatch);
-	}
-	break;
-      case T_FILE:
-	if (RFILE(obj)->fptr) {
-	    size += rb_io_memsize(RFILE(obj)->fptr);
-	}
-	break;
-      case T_RATIONAL:
-      case T_COMPLEX:
-	break;
-      case T_ICLASS:
-	/* iClass shares table with the module */
-	break;
-
-      case T_FLOAT:
-	break;
-
-      case T_BIGNUM:
-	if (!(RBASIC(obj)->flags & RBIGNUM_EMBED_FLAG) && RBIGNUM_DIGITS(obj)) {
-	    size += RBIGNUM_LEN(obj) * sizeof(BDIGIT);
-	}
-	break;
-      case T_NODE:
-	switch (nd_type(obj)) {
-	  case NODE_SCOPE:
-	    if (RNODE(obj)->u1.tbl) {
-		/* TODO: xfree(RANY(obj)->as.node.u1.tbl); */
-	    }
-	    break;
-	  case NODE_ALLOCA:
-	    /* TODO: xfree(RANY(obj)->as.node.u1.node); */
-	    ;
-	}
-	break;			/* no need to free iv_tbl */
-
-      case T_STRUCT:
-	if ((RBASIC(obj)->flags & RSTRUCT_EMBED_LEN_MASK) == 0 &&
-	    RSTRUCT(obj)->as.heap.ptr) {
-	    size += sizeof(VALUE) * RSTRUCT_LEN(obj);
-	}
-	break;
-
-      case T_ZOMBIE:
-	break;
-
-      default:
-	rb_bug("objspace/memsize_of(): unknown data type 0x%x(%p)",
-	       BUILTIN_TYPE(obj), (void*)obj);
-    }
-
-    return size;
-}
+#include "internal/class.h"
+#include "internal/compilers.h"
+#include "internal/hash.h"
+#include "internal/imemo.h"
+#include "node.h"
+#include "ruby/io.h"
+#include "ruby/re.h"
+#include "ruby/st.h"
+#include "symbol.h"
 
 /*
  *  call-seq:
  *    ObjectSpace.memsize_of(obj) -> Integer
  *
- *  Return consuming memory size of obj.
+ *  Return consuming memory size of obj in bytes.
  *
- *  Note that the return size is incomplete.  You need to deal with
- *  this information as only a *HINT*.  Especially, the size of
- *  T_DATA may not be correct.
+ *  Note that the return size is incomplete.  You need to deal with this
+ *  information as only a *HINT*. Especially, the size of +T_DATA+ may not be
+ *  correct.
  *
- *  This method is not expected to work except C Ruby.
+ *  This method is only expected to work with C Ruby.
+ *
+ *  From Ruby 2.2, memsize_of(obj) returns a memory size includes
+ *  sizeof(RVALUE).
  */
 
 static VALUE
 memsize_of_m(VALUE self, VALUE obj)
 {
-    return SIZET2NUM(memsize_of(obj));
+    return SIZET2NUM(rb_obj_memsize_of(obj));
 }
 
 struct total_data {
@@ -178,16 +61,14 @@ total_i(void *vstart, void *vend, size_t stride, void *ptr)
 	if (RBASIC(v)->flags) {
 	    switch (BUILTIN_TYPE(v)) {
 	      case T_NONE:
+	      case T_IMEMO:
 	      case T_ICLASS:
 	      case T_NODE:
 	      case T_ZOMBIE:
 		continue;
-	      case T_CLASS:
-		if (FL_TEST(v, FL_SINGLETON))
-		  continue;
 	      default:
 		if (data->klass == 0 || rb_obj_is_kind_of(v, data->klass)) {
-		    data->total += memsize_of(v);
+		    data->total += rb_obj_memsize_of(v);
 		}
 	    }
 	}
@@ -200,27 +81,28 @@ total_i(void *vstart, void *vend, size_t stride, void *ptr)
  *  call-seq:
  *    ObjectSpace.memsize_of_all([klass]) -> Integer
  *
- *  Return consuming memory size of all living objects.
- *  If klass (should be Class object) is given, return the total
- *  memory size of instances of the given class.
+ *  Return consuming memory size of all living objects in bytes.
  *
- *  Note that the returned size is incomplete.  You need to deal with
- *  this information as only a *HINT*.  Especially, the size of
- *  T_DATA may not be correct.
+ *  If +klass+ (should be Class object) is given, return the total memory size
+ *  of instances of the given class.
+ *
+ *  Note that the returned size is incomplete. You need to deal with this
+ *  information as only a *HINT*. Especially, the size of +T_DATA+ may not be
+ *  correct.
  *
  *  Note that this method does *NOT* return total malloc'ed memory size.
  *
  *  This method can be defined by the following Ruby code:
  *
- *  def memsize_of_all klass = false
- *    total = 0
- *    ObjectSpace.each_object{|e|
- *      total += ObjectSpace.memsize_of(e) if klass == false || e.kind_of?(klass)
- *    }
- *    total
- *  end
+ *	def memsize_of_all klass = false
+ *  	  total = 0
+ *  	  ObjectSpace.each_object{|e|
+ *  	    total += ObjectSpace.memsize_of(e) if klass == false || e.kind_of?(klass)
+ *  	  }
+ *  	  total
+ *  	end
  *
- *  This method is not expected to work except C Ruby.
+ *  This method is only expected to work with C Ruby.
  */
 
 static VALUE
@@ -245,6 +127,27 @@ set_zero_i(st_data_t key, st_data_t val, st_data_t arg)
     return ST_CONTINUE;
 }
 
+static VALUE
+setup_hash(int argc, VALUE *argv)
+{
+    VALUE hash;
+
+    if (rb_scan_args(argc, argv, "01", &hash) == 1) {
+        if (!RB_TYPE_P(hash, T_HASH))
+            rb_raise(rb_eTypeError, "non-hash given");
+    }
+
+    if (hash == Qnil) {
+        hash = rb_hash_new();
+    }
+    else if (!RHASH_EMPTY_P(hash)) {
+        /* WB: no new reference */
+        st_foreach(RHASH_TBL_RAW(hash), set_zero_i, hash);
+    }
+
+    return hash;
+}
+
 static int
 cos_i(void *vstart, void *vend, size_t stride, void *data)
 {
@@ -253,7 +156,7 @@ cos_i(void *vstart, void *vend, size_t stride, void *data)
 
     for (;v != (VALUE)vend; v += stride) {
 	if (RBASIC(v)->flags) {
-	    counts[BUILTIN_TYPE(v)] += memsize_of(v);
+	    counts[BUILTIN_TYPE(v)] += rb_obj_memsize_of(v);
 	}
     }
     return 0;
@@ -287,8 +190,10 @@ type2sym(enum ruby_value_type i)
 	CASE_TYPE(T_SYMBOL);
 	CASE_TYPE(T_FIXNUM);
 	CASE_TYPE(T_UNDEF);
+	CASE_TYPE(T_IMEMO);
 	CASE_TYPE(T_NODE);
 	CASE_TYPE(T_ICLASS);
+        CASE_TYPE(T_MOVED);
 	CASE_TYPE(T_ZOMBIE);
 #undef CASE_TYPE
       default: rb_bug("type2sym: unknown type (%d)", i);
@@ -304,7 +209,7 @@ type2sym(enum ruby_value_type i)
  *
  *  Note that this information is incomplete.  You need to deal with
  *  this information as only a *HINT*.  Especially, total size of
- *  T_DATA may not right size.
+ *  T_DATA may be wrong.
  *
  *  It returns a hash as:
  *    {:TOTAL=>1461154, :T_CLASS=>158280, :T_MODULE=>20672, :T_STRING=>527249, ...}
@@ -316,7 +221,7 @@ type2sym(enum ruby_value_type i)
  *  The contents of the returned hash is implementation defined.
  *  It may be changed in future.
  *
- *  This method is not expected to work except C Ruby.
+ *  This method is only expected to work with C Ruby.
  */
 
 static VALUE
@@ -325,12 +230,7 @@ count_objects_size(int argc, VALUE *argv, VALUE os)
     size_t counts[T_MASK+1];
     size_t total = 0;
     enum ruby_value_type i;
-    VALUE hash;
-
-    if (rb_scan_args(argc, argv, "01", &hash) == 1) {
-        if (!RB_TYPE_P(hash, T_HASH))
-            rb_raise(rb_eTypeError, "non-hash given");
-    }
+    VALUE hash = setup_hash(argc, argv);
 
     for (i = 0; i <= T_MASK; i++) {
 	counts[i] = 0;
@@ -338,13 +238,6 @@ count_objects_size(int argc, VALUE *argv, VALUE os)
 
     rb_objspace_each_objects(cos_i, &counts[0]);
 
-    if (hash == Qnil) {
-        hash = rb_hash_new();
-    }
-    else if (!RHASH_EMPTY_P(hash)) {
-        st_foreach(RHASH_TBL(hash), set_zero_i, hash);
-    }
-
     for (i = 0; i <= T_MASK; i++) {
 	if (counts[i]) {
 	    VALUE type = type2sym(i);
@@ -356,6 +249,77 @@ count_objects_size(int argc, VALUE *argv, VALUE os)
     return hash;
 }
 
+struct dynamic_symbol_counts {
+    size_t mortal;
+    size_t immortal;
+};
+
+static int
+cs_i(void *vstart, void *vend, size_t stride, void *n)
+{
+    struct dynamic_symbol_counts *counts = (struct dynamic_symbol_counts *)n;
+    VALUE v = (VALUE)vstart;
+
+    for (; v != (VALUE)vend; v += stride) {
+	if (RBASIC(v)->flags && BUILTIN_TYPE(v) == T_SYMBOL) {
+	    ID id = RSYMBOL(v)->id;
+	    if ((id & ~ID_SCOPE_MASK) == 0) {
+		counts->mortal++;
+	    }
+	    else {
+		counts->immortal++;
+	    }
+	}
+    }
+
+    return 0;
+}
+
+size_t rb_sym_immortal_count(void);
+
+/*
+ *  call-seq:
+ *     ObjectSpace.count_symbols([result_hash]) -> hash
+ *
+ *  Counts symbols for each Symbol type.
+ *
+ *  This method is only for MRI developers interested in performance and memory
+ *  usage of Ruby programs.
+ *
+ *  If the optional argument, result_hash, is given, it is overwritten and
+ *  returned. This is intended to avoid probe effect.
+ *
+ *  Note:
+ *  The contents of the returned hash is implementation defined.
+ *  It may be changed in future.
+ *
+ *  This method is only expected to work with C Ruby.
+ *
+ *  On this version of MRI, they have 3 types of Symbols (and 1 total counts).
+ *
+ *   * mortal_dynamic_symbol: GC target symbols (collected by GC)
+ *   * immortal_dynamic_symbol: Immortal symbols promoted from dynamic symbols (do not collected by GC)
+ *   * immortal_static_symbol: Immortal symbols (do not collected by GC)
+ *   * immortal_symbol: total immortal symbols (immortal_dynamic_symbol+immortal_static_symbol)
+ */
+
+static VALUE
+count_symbols(int argc, VALUE *argv, VALUE os)
+{
+    struct dynamic_symbol_counts dynamic_counts = {0, 0};
+    VALUE hash = setup_hash(argc, argv);
+
+    size_t immortal_symbols = rb_sym_immortal_count();
+    rb_objspace_each_objects(cs_i, &dynamic_counts);
+
+    rb_hash_aset(hash, ID2SYM(rb_intern("mortal_dynamic_symbol")),   SIZET2NUM(dynamic_counts.mortal));
+    rb_hash_aset(hash, ID2SYM(rb_intern("immortal_dynamic_symbol")), SIZET2NUM(dynamic_counts.immortal));
+    rb_hash_aset(hash, ID2SYM(rb_intern("immortal_static_symbol")),  SIZET2NUM(immortal_symbols - dynamic_counts.immortal));
+    rb_hash_aset(hash, ID2SYM(rb_intern("immortal_symbol")),         SIZET2NUM(immortal_symbols));
+
+    return hash;
+}
+
 static int
 cn_i(void *vstart, void *vend, size_t stride, void *n)
 {
@@ -378,33 +342,29 @@ cn_i(void *vstart, void *vend, size_t stride, void *n)
  *
  *  Counts nodes for each node type.
  *
- *  This method is not for ordinary Ruby programmers, but for MRI developers
- *  who have interest in MRI performance and memory usage.
+ *  This method is only for MRI developers interested in performance and memory
+ *  usage of Ruby programs.
  *
  *  It returns a hash as:
- *  {:NODE_METHOD=>2027, :NODE_FBODY=>1927, :NODE_CFUNC=>1798, ...}
  *
- *  If the optional argument, result_hash, is given,
- *  it is overwritten and returned.
- *  This is intended to avoid probe effect.
+ *	{:NODE_METHOD=>2027, :NODE_FBODY=>1927, :NODE_CFUNC=>1798, ...}
  *
+ *  If the optional argument, result_hash, is given, it is overwritten and
+ *  returned. This is intended to avoid probe effect.
+ *
+ *  Note:
  *  The contents of the returned hash is implementation defined.
  *  It may be changed in future.
  *
- *  This method is not expected to work except C Ruby.
+ *  This method is only expected to work with C Ruby.
  */
 
 static VALUE
 count_nodes(int argc, VALUE *argv, VALUE os)
 {
     size_t nodes[NODE_LAST+1];
-    size_t i;
-    VALUE hash;
-
-    if (rb_scan_args(argc, argv, "01", &hash) == 1) {
-        if (!RB_TYPE_P(hash, T_HASH))
-            rb_raise(rb_eTypeError, "non-hash given");
-    }
+    enum node_type i;
+    VALUE hash = setup_hash(argc, argv);
 
     for (i = 0; i <= NODE_LAST; i++) {
 	nodes[i] = 0;
@@ -412,28 +372,25 @@ count_nodes(int argc, VALUE *argv, VALUE os)
 
     rb_objspace_each_objects(cn_i, &nodes[0]);
 
-    if (hash == Qnil) {
-        hash = rb_hash_new();
-    }
-    else if (!RHASH_EMPTY_P(hash)) {
-        st_foreach(RHASH_TBL(hash), set_zero_i, hash);
-    }
-
     for (i=0; i<NODE_LAST; i++) {
 	if (nodes[i] != 0) {
 	    VALUE node;
 	    switch (i) {
-#define COUNT_NODE(n) case n: node = ID2SYM(rb_intern(#n)); break;
+#define COUNT_NODE(n) case n: node = ID2SYM(rb_intern(#n)); goto set
 		COUNT_NODE(NODE_SCOPE);
 		COUNT_NODE(NODE_BLOCK);
 		COUNT_NODE(NODE_IF);
+		COUNT_NODE(NODE_UNLESS);
 		COUNT_NODE(NODE_CASE);
+		COUNT_NODE(NODE_CASE2);
+		COUNT_NODE(NODE_CASE3);
 		COUNT_NODE(NODE_WHEN);
-		COUNT_NODE(NODE_OPT_N);
+		COUNT_NODE(NODE_IN);
 		COUNT_NODE(NODE_WHILE);
 		COUNT_NODE(NODE_UNTIL);
 		COUNT_NODE(NODE_ITER);
 		COUNT_NODE(NODE_FOR);
+		COUNT_NODE(NODE_FOR_MASGN);
 		COUNT_NODE(NODE_BREAK);
 		COUNT_NODE(NODE_NEXT);
 		COUNT_NODE(NODE_REDO);
@@ -450,22 +407,22 @@ count_nodes(int argc, VALUE *argv, VALUE os)
 		COUNT_NODE(NODE_DASGN_CURR);
 		COUNT_NODE(NODE_GASGN);
 		COUNT_NODE(NODE_IASGN);
-		COUNT_NODE(NODE_IASGN2);
 		COUNT_NODE(NODE_CDECL);
 		COUNT_NODE(NODE_CVASGN);
-		COUNT_NODE(NODE_CVDECL);
 		COUNT_NODE(NODE_OP_ASGN1);
 		COUNT_NODE(NODE_OP_ASGN2);
 		COUNT_NODE(NODE_OP_ASGN_AND);
 		COUNT_NODE(NODE_OP_ASGN_OR);
 		COUNT_NODE(NODE_OP_CDECL);
 		COUNT_NODE(NODE_CALL);
+		COUNT_NODE(NODE_OPCALL);
 		COUNT_NODE(NODE_FCALL);
 		COUNT_NODE(NODE_VCALL);
+		COUNT_NODE(NODE_QCALL);
 		COUNT_NODE(NODE_SUPER);
 		COUNT_NODE(NODE_ZSUPER);
-		COUNT_NODE(NODE_ARRAY);
-		COUNT_NODE(NODE_ZARRAY);
+		COUNT_NODE(NODE_LIST);
+		COUNT_NODE(NODE_ZLIST);
 		COUNT_NODE(NODE_VALUES);
 		COUNT_NODE(NODE_HASH);
 		COUNT_NODE(NODE_RETURN);
@@ -488,7 +445,7 @@ count_nodes(int argc, VALUE *argv, VALUE os)
 		COUNT_NODE(NODE_DXSTR);
 		COUNT_NODE(NODE_EVSTR);
 		COUNT_NODE(NODE_DREGX);
-		COUNT_NODE(NODE_DREGX_ONCE);
+		COUNT_NODE(NODE_ONCE);
 		COUNT_NODE(NODE_ARGS);
 		COUNT_NODE(NODE_ARGS_AUX);
 		COUNT_NODE(NODE_OPT_ARG);
@@ -497,8 +454,6 @@ count_nodes(int argc, VALUE *argv, VALUE os)
 		COUNT_NODE(NODE_ARGSCAT);
 		COUNT_NODE(NODE_ARGSPUSH);
 		COUNT_NODE(NODE_SPLAT);
-		COUNT_NODE(NODE_TO_ARY);
-		COUNT_NODE(NODE_BLOCK_ARG);
 		COUNT_NODE(NODE_BLOCK_PASS);
 		COUNT_NODE(NODE_DEFN);
 		COUNT_NODE(NODE_DEFS);
@@ -510,7 +465,6 @@ count_nodes(int argc, VALUE *argv, VALUE os)
 		COUNT_NODE(NODE_SCLASS);
 		COUNT_NODE(NODE_COLON2);
 		COUNT_NODE(NODE_COLON3);
-		COUNT_NODE(NODE_CREF);
 		COUNT_NODE(NODE_DOT2);
 		COUNT_NODE(NODE_DOT3);
 		COUNT_NODE(NODE_FLIP2);
@@ -522,17 +476,17 @@ count_nodes(int argc, VALUE *argv, VALUE os)
 		COUNT_NODE(NODE_ERRINFO);
 		COUNT_NODE(NODE_DEFINED);
 		COUNT_NODE(NODE_POSTEXE);
-		COUNT_NODE(NODE_ALLOCA);
-		COUNT_NODE(NODE_BMETHOD);
-		COUNT_NODE(NODE_MEMO);
-		COUNT_NODE(NODE_IFUNC);
 		COUNT_NODE(NODE_DSYM);
 		COUNT_NODE(NODE_ATTRASGN);
-		COUNT_NODE(NODE_PRELUDE);
 		COUNT_NODE(NODE_LAMBDA);
+		COUNT_NODE(NODE_ARYPTN);
+		COUNT_NODE(NODE_FNDPTN);
+		COUNT_NODE(NODE_HSHPTN);
 #undef COUNT_NODE
-	      default: node = INT2FIX(i);
+	      case NODE_LAST: break;
 	    }
+	    UNREACHABLE;
+	  set:
 	    rb_hash_aset(hash, node, SIZET2NUM(nodes[i]));
 	}
     }
@@ -575,51 +529,121 @@ cto_i(void *vstart, void *vend, size_t stride, void *data)
  *  call-seq:
  *     ObjectSpace.count_tdata_objects([result_hash]) -> hash
  *
- *  Counts objects for each T_DATA type.
+ *  Counts objects for each +T_DATA+ type.
  *
- *  This method is not for ordinary Ruby programmers, but for MRI developers
- *  who interest on MRI performance.
+ *  This method is only for MRI developers interested in performance and memory
+ *  usage of Ruby programs.
  *
  *  It returns a hash as:
- *  {RubyVM::InstructionSequence=>504, :parser=>5, :barrier=>6,
- *   :mutex=>6, Proc=>60, RubyVM::Env=>57, Mutex=>1, Encoding=>99,
- *   ThreadGroup=>1, Binding=>1, Thread=>1, RubyVM=>1, :iseq=>1,
- *   Random=>1, ARGF.class=>1, Data=>1, :autoload=>3, Time=>2}
- *  # T_DATA objects existing at startup on r32276.
  *
- *  If the optional argument, result_hash, is given,
- *  it is overwritten and returned.
- *  This is intended to avoid probe effect.
+ *	{RubyVM::InstructionSequence=>504, :parser=>5, :barrier=>6,
+ *  	 :mutex=>6, Proc=>60, RubyVM::Env=>57, Mutex=>1, Encoding=>99,
+ *  	 ThreadGroup=>1, Binding=>1, Thread=>1, RubyVM=>1, :iseq=>1,
+ *  	 Random=>1, ARGF.class=>1, Data=>1, :autoload=>3, Time=>2}
+ *  	# T_DATA objects existing at startup on r32276.
  *
- *  The contents of the returned hash is implementation defined.
- *  It may be changed in future.
+ *  If the optional argument, result_hash, is given, it is overwritten and
+ *  returned. This is intended to avoid probe effect.
+ *
+ *  The contents of the returned hash is implementation specific and may change
+ *  in the future.
  *
  *  In this version, keys are Class object or Symbol object.
+ *
  *  If object is kind of normal (accessible) object, the key is Class object.
  *  If object is not a kind of normal (internal) object, the key is symbol
  *  name, registered by rb_data_type_struct.
  *
- *  This method is not expected to work except C Ruby.
+ *  This method is only expected to work with C Ruby.
  */
 
 static VALUE
 count_tdata_objects(int argc, VALUE *argv, VALUE self)
 {
-    VALUE hash;
-
-    if (rb_scan_args(argc, argv, "01", &hash) == 1) {
-        if (!RB_TYPE_P(hash, T_HASH))
-            rb_raise(rb_eTypeError, "non-hash given");
-    }
-
-    if (hash == Qnil) {
-        hash = rb_hash_new();
-    }
-    else if (!RHASH_EMPTY_P(hash)) {
-        st_foreach(RHASH_TBL(hash), set_zero_i, hash);
-    }
-
+    VALUE hash = setup_hash(argc, argv);
     rb_objspace_each_objects(cto_i, (void *)hash);
+    return hash;
+}
+
+static ID imemo_type_ids[IMEMO_MASK+1];
+
+static int
+count_imemo_objects_i(void *vstart, void *vend, size_t stride, void *data)
+{
+    VALUE hash = (VALUE)data;
+    VALUE v = (VALUE)vstart;
+
+    for (; v != (VALUE)vend; v += stride) {
+	if (RBASIC(v)->flags && BUILTIN_TYPE(v) == T_IMEMO) {
+	    VALUE counter;
+	    VALUE key = ID2SYM(imemo_type_ids[imemo_type(v)]);
+
+	    counter = rb_hash_aref(hash, key);
+
+	    if (NIL_P(counter)) {
+		counter = INT2FIX(1);
+	    }
+	    else {
+		counter = INT2FIX(FIX2INT(counter) + 1);
+	    }
+
+	    rb_hash_aset(hash, key, counter);
+	}
+    }
+
+    return 0;
+}
+
+/*
+ *  call-seq:
+ *     ObjectSpace.count_imemo_objects([result_hash]) -> hash
+ *
+ *  Counts objects for each +T_IMEMO+ type.
+ *
+ *  This method is only for MRI developers interested in performance and memory
+ *  usage of Ruby programs.
+ *
+ *  It returns a hash as:
+ *
+ *       {:imemo_ifunc=>8,
+ *        :imemo_svar=>7,
+ *        :imemo_cref=>509,
+ *        :imemo_memo=>1,
+ *        :imemo_throw_data=>1}
+ *
+ *  If the optional argument, result_hash, is given, it is overwritten and
+ *  returned. This is intended to avoid probe effect.
+ *
+ *  The contents of the returned hash is implementation specific and may change
+ *  in the future.
+ *
+ *  In this version, keys are symbol objects.
+ *
+ *  This method is only expected to work with C Ruby.
+ */
+
+static VALUE
+count_imemo_objects(int argc, VALUE *argv, VALUE self)
+{
+    VALUE hash = setup_hash(argc, argv);
+
+    if (imemo_type_ids[0] == 0) {
+        imemo_type_ids[0] = rb_intern("imemo_env");
+	imemo_type_ids[1] = rb_intern("imemo_cref");
+	imemo_type_ids[2] = rb_intern("imemo_svar");
+	imemo_type_ids[3] = rb_intern("imemo_throw_data");
+	imemo_type_ids[4] = rb_intern("imemo_ifunc");
+	imemo_type_ids[5] = rb_intern("imemo_memo");
+	imemo_type_ids[6] = rb_intern("imemo_ment");
+	imemo_type_ids[7] = rb_intern("imemo_iseq");
+	imemo_type_ids[8] = rb_intern("imemo_tmpbuf");
+        imemo_type_ids[9] = rb_intern("imemo_ast");
+        imemo_type_ids[10] = rb_intern("imemo_parser_strterm");
+        imemo_type_ids[11] = rb_intern("imemo_callinfo");
+        imemo_type_ids[12] = rb_intern("imemo_callcache");
+    }
+
+    rb_objspace_each_objects(count_imemo_objects_i, (void *)hash);
 
     return hash;
 }
@@ -634,20 +658,21 @@ static size_t
 iow_size(const void *ptr)
 {
     VALUE obj = (VALUE)ptr;
-    return memsize_of(obj);
+    return rb_obj_memsize_of(obj);
 }
 
 static const rb_data_type_t iow_data_type = {
     "ObjectSpace::InternalObjectWrapper",
     {iow_mark, 0, iow_size,},
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY
 };
 
-static VALUE rb_mInternalObjectWrapper;
+static VALUE rb_cInternalObjectWrapper;
 
 static VALUE
 iow_newobj(VALUE obj)
 {
-    return rb_data_typed_object_alloc(rb_mInternalObjectWrapper, (void *)obj, &iow_data_type);
+    return TypedData_Wrap_Struct(rb_cInternalObjectWrapper, &iow_data_type, (void *)obj);
 }
 
 /* Returns the type of the internal object. */
@@ -665,7 +690,7 @@ iow_inspect(VALUE self)
     VALUE obj = (VALUE)DATA_PTR(self);
     VALUE type = type2sym(BUILTIN_TYPE(obj));
 
-    return rb_sprintf("#<InternalObject:%p %s>", (void *)obj, rb_id2name(SYM2ID(type)));
+    return rb_sprintf("#<InternalObject:%p %"PRIsVALUE">", (void *)obj, rb_sym2str(type));
 }
 
 /* Returns the Object#object_id of the internal object. */
@@ -712,24 +737,25 @@ collect_values(st_data_t key, st_data_t value, st_data_t data)
  *  [MRI specific feature] Return all reachable objects from `obj'.
  *
  *  This method returns all reachable objects from `obj'.
- *  If `obj' has two or more references to the same object `x',
- *  then returned array only includes one `x' object.
- *  If `obj' is a non-markable (non-heap management) object such as
- *  true, false, nil, symbols and Fixnums (and Flonum) then it simply
- *  returns nil.
  *
- *  If `obj' has references to an internal object, then it returns
- *  instances of `ObjectSpace::InternalObjectWrapper' class.
- *  This object contains a reference to an internal object and
- *  you can check the type of internal object with `type' method.
+ *  If `obj' has two or more references to the same object `x', then returned
+ *  array only includes one `x' object.
  *
- *  If `obj' is instance of `ObjectSpace::InternalObjectWrapper'
- *  class, then this method returns all reachable object from
- *  an internal object, which is pointed by `obj'.
+ *  If `obj' is a non-markable (non-heap management) object such as true,
+ *  false, nil, symbols and Fixnums (and Flonum) then it simply returns nil.
+ *
+ *  If `obj' has references to an internal object, then it returns instances of
+ *  ObjectSpace::InternalObjectWrapper class. This object contains a reference
+ *  to an internal object and you can check the type of internal object with
+ *  `type' method.
+ *
+ *  If `obj' is instance of ObjectSpace::InternalObjectWrapper class, then this
+ *  method returns all reachable object from an internal object, which is
+ *  pointed by `obj'.
  *
  *  With this method, you can find memory leaks.
  *
- *  This method is not expected to work except in C Ruby.
+ *  This method is only expected to work except with C Ruby.
  *
  *  Example:
  *    ObjectSpace.reachable_objects_from(['a', 'b', 'c'])
@@ -770,14 +796,146 @@ reachable_objects_from(VALUE self, VALUE obj)
     }
 }
 
+struct rofr_data {
+    VALUE categories;
+    const char *last_category;
+    VALUE last_category_str;
+    VALUE last_category_objects;
+};
+
+static void
+reachable_object_from_root_i(const char *category, VALUE obj, void *ptr)
+{
+    struct rofr_data *data = (struct rofr_data *)ptr;
+    VALUE category_str;
+    VALUE category_objects;
+
+    if (category == data->last_category) {
+	category_str = data->last_category_str;
+	category_objects = data->last_category_objects;
+    }
+    else {
+	data->last_category = category;
+	category_str = data->last_category_str = rb_str_new2(category);
+	category_objects = data->last_category_objects = rb_ident_hash_new();
+	if (!NIL_P(rb_hash_lookup(data->categories, category_str))) {
+	    rb_bug("reachable_object_from_root_i: category should insert at once");
+	}
+	rb_hash_aset(data->categories, category_str, category_objects);
+    }
+
+    if (rb_objspace_markable_object_p(obj) &&
+	obj != data->categories &&
+	obj != data->last_category_objects) {
+	if (rb_objspace_internal_object_p(obj)) {
+	    obj = iow_newobj(obj);
+	}
+	rb_hash_aset(category_objects, obj, obj);
+    }
+}
+
+static int
+collect_values_of_values(VALUE category, VALUE category_objects, VALUE categories)
+{
+    VALUE ary = rb_ary_new();
+    rb_hash_foreach(category_objects, collect_values, ary);
+    rb_hash_aset(categories, category, ary);
+    return ST_CONTINUE;
+}
+
+/*
+ *  call-seq:
+ *     ObjectSpace.reachable_objects_from_root -> hash
+ *
+ *  [MRI specific feature] Return all reachable objects from root.
+ */
+static VALUE
+reachable_objects_from_root(VALUE self)
+{
+    struct rofr_data data;
+    VALUE hash = data.categories = rb_ident_hash_new();
+    data.last_category = 0;
+
+    rb_objspace_reachable_objects_from_root(reachable_object_from_root_i, &data);
+    rb_hash_foreach(hash, collect_values_of_values, hash);
+
+    return hash;
+}
+
+static VALUE
+wrap_klass_iow(VALUE klass)
+{
+    if (!RTEST(klass)) {
+	return Qnil;
+    }
+    else if (RB_TYPE_P(klass, T_ICLASS) ||
+             CLASS_OF(klass) == Qfalse /* hidden object */) {
+	return iow_newobj(klass);
+    }
+    else {
+	return klass;
+    }
+}
+
+/*
+ *  call-seq:
+ *     ObjectSpace.internal_class_of(obj) -> Class or Module
+ *
+ *  [MRI specific feature] Return internal class of obj.
+ *  obj can be an instance of InternalObjectWrapper.
+ *
+ *  Note that you should not use this method in your application.
+ */
+static VALUE
+objspace_internal_class_of(VALUE self, VALUE obj)
+{
+    VALUE klass;
+
+    if (rb_typeddata_is_kind_of(obj, &iow_data_type)) {
+	obj = (VALUE)DATA_PTR(obj);
+    }
+
+    klass = CLASS_OF(obj);
+    return wrap_klass_iow(klass);
+}
+
+/*
+ *  call-seq:
+ *     ObjectSpace.internal_super_of(cls) -> Class or Module
+ *
+ *  [MRI specific feature] Return internal super class of cls (Class or Module).
+ *  obj can be an instance of InternalObjectWrapper.
+ *
+ *  Note that you should not use this method in your application.
+ */
+static VALUE
+objspace_internal_super_of(VALUE self, VALUE obj)
+{
+    VALUE super;
+
+    if (rb_typeddata_is_kind_of(obj, &iow_data_type)) {
+	obj = (VALUE)DATA_PTR(obj);
+    }
+
+    switch (OBJ_BUILTIN_TYPE(obj)) {
+      case T_MODULE:
+      case T_CLASS:
+      case T_ICLASS:
+	super = RCLASS_SUPER(obj);
+	break;
+      default:
+	rb_raise(rb_eArgError, "class or module is expected");
+    }
+
+    return wrap_klass_iow(super);
+}
+
 void Init_object_tracing(VALUE rb_mObjSpace);
-void Init_gc_hook(VALUE rb_mObjSpace);
+void Init_objspace_dump(VALUE rb_mObjSpace);
 
 /*
  * Document-module: ObjectSpace
  *
- * == The objspace library
- *
  * The objspace library extends the ObjectSpace module and adds several
  * methods to get internal statistic information about
  * object/memory management.
@@ -793,6 +951,7 @@ void Init_gc_hook(VALUE rb_mObjSpace);
 void
 Init_objspace(void)
 {
+#undef rb_intern
     VALUE rb_mObjSpace;
 #if 0
     rb_mObjSpace = rb_define_module("ObjectSpace"); /* let rdoc know */
@@ -803,10 +962,16 @@ Init_objspace(void)
     rb_define_module_function(rb_mObjSpace, "memsize_of_all", memsize_of_all_m, -1);
 
     rb_define_module_function(rb_mObjSpace, "count_objects_size", count_objects_size, -1);
+    rb_define_module_function(rb_mObjSpace, "count_symbols", count_symbols, -1);
     rb_define_module_function(rb_mObjSpace, "count_nodes", count_nodes, -1);
     rb_define_module_function(rb_mObjSpace, "count_tdata_objects", count_tdata_objects, -1);
+    rb_define_module_function(rb_mObjSpace, "count_imemo_objects", count_imemo_objects, -1);
 
     rb_define_module_function(rb_mObjSpace, "reachable_objects_from", reachable_objects_from, 1);
+    rb_define_module_function(rb_mObjSpace, "reachable_objects_from_root", reachable_objects_from_root, 0);
+
+    rb_define_module_function(rb_mObjSpace, "internal_class_of", objspace_internal_class_of, 1);
+    rb_define_module_function(rb_mObjSpace, "internal_super_of", objspace_internal_super_of, 1);
 
     /*
      * This class is used as a return value from
@@ -817,11 +982,11 @@ Init_objspace(void)
      *
      * You can use the #type method to check the type of the internal object.
      */
-    rb_mInternalObjectWrapper = rb_define_class_under(rb_mObjSpace, "InternalObjectWrapper", rb_cObject);
-    rb_define_method(rb_mInternalObjectWrapper, "type", iow_type, 0);
-    rb_define_method(rb_mInternalObjectWrapper, "inspect", iow_inspect, 0);
-    rb_define_method(rb_mInternalObjectWrapper, "internal_object_id", iow_internal_object_id, 0);
+    rb_cInternalObjectWrapper = rb_define_class_under(rb_mObjSpace, "InternalObjectWrapper", rb_cObject);
+    rb_define_method(rb_cInternalObjectWrapper, "type", iow_type, 0);
+    rb_define_method(rb_cInternalObjectWrapper, "inspect", iow_inspect, 0);
+    rb_define_method(rb_cInternalObjectWrapper, "internal_object_id", iow_internal_object_id, 0);
 
     Init_object_tracing(rb_mObjSpace);
-    Init_gc_hook(rb_mObjSpace);
+    Init_objspace_dump(rb_mObjSpace);
 }
