@@ -1,5 +1,4 @@
 # frozen_string_literal: true
-require "spec_helper"
 
 RSpec.describe "bundle install with gems on multiple sources" do
   # repo1 is built automatically before all of the specs run
@@ -21,28 +20,25 @@ RSpec.describe "bundle install with gems on multiple sources" do
 
       before do
         gemfile <<-G
-          source "file://#{gem_repo3}"
-          source "file://#{gem_repo1}"
+          source "#{file_uri_for(gem_repo3)}"
+          source "#{file_uri_for(gem_repo1)}"
           gem "rack-obama"
           gem "rack"
         G
-        bundle "config major_deprecations true"
       end
 
-      it "warns about ambiguous gems, but installs anyway, prioritizing sources last to first" do
+      it "warns about ambiguous gems, but installs anyway, prioritizing sources last to first", :bundler => "2" do
         bundle :install
 
-        expect(out).to have_major_deprecation a_string_including("Your Gemfile contains multiple primary sources.")
-        expect(out).to include("Warning: the gem 'rack' was found in multiple sources.")
-        expect(out).to include("Installed from: file:#{gem_repo1}")
+        expect(err).to include("Warning: the gem 'rack' was found in multiple sources.")
+        expect(err).to include("Installed from: #{file_uri_for(gem_repo1)}")
         expect(the_bundle).to include_gems("rack-obama 1.0.0", "rack 1.0.0", :source => "remote1")
       end
 
-      it "errors when disable_multisource is set" do
-        bundle "config disable_multisource true"
-        bundle :install
-        expect(out).to include("Each source after the first must include a block")
-        expect(exitstatus).to eq(4) if exitstatus
+      it "fails", :bundler => "3" do
+        bundle :instal, :raise_on_error => false
+        expect(err).to include("Each source after the first must include a block")
+        expect(exitstatus).to eq(4)
       end
     end
 
@@ -51,22 +47,25 @@ RSpec.describe "bundle install with gems on multiple sources" do
 
       before do
         gemfile <<-G
-          source "file://#{gem_repo3}"
-          source "file://#{gem_repo1}"
+          source "#{file_uri_for(gem_repo3)}"
+          source "#{file_uri_for(gem_repo1)}"
           gem "rack-obama"
           gem "rack", "1.0.0" # force it to install the working version in repo1
         G
-        bundle "config major_deprecations true"
       end
 
-      it "warns about ambiguous gems, but installs anyway" do
+      it "warns about ambiguous gems, but installs anyway", :bundler => "2" do
         bundle :install
-
-        expect(out).to have_major_deprecation a_string_including("Your Gemfile contains multiple primary sources.")
-        expect(out).to include("Warning: the gem 'rack' was found in multiple sources.")
-        expect(out).to include("Installed from: file:#{gem_repo1}")
+        expect(err).to include("Warning: the gem 'rack' was found in multiple sources.")
+        expect(err).to include("Installed from: #{file_uri_for(gem_repo1)}")
         expect(the_bundle).to include_gems("rack-obama 1.0.0", "rack 1.0.0", :source => "remote1")
       end
+
+      it "fails", :bundler => "3" do
+        bundle :install, :raise_on_error => false
+        expect(err).to include("Each source after the first must include a block")
+        expect(exitstatus).to eq(4)
+      end
     end
   end
 
@@ -79,11 +78,15 @@ RSpec.describe "bundle install with gems on multiple sources" do
           build_gem "rack", "1.0.0" do |s|
             s.write "lib/rack.rb", "RACK = 'FAIL'"
           end
+
+          build_gem "rack-obama" do |s|
+            s.add_dependency "rack"
+          end
         end
 
         gemfile <<-G
-          source "file://#{gem_repo3}"
-          source "file://#{gem_repo1}" do
+          source "#{file_uri_for(gem_repo3)}"
+          source "#{file_uri_for(gem_repo1)}" do
             gem "thin" # comes first to test name sorting
             gem "rack"
           end
@@ -99,14 +102,14 @@ RSpec.describe "bundle install with gems on multiple sources" do
       end
 
       it "can cache and deploy" do
-        bundle :package
+        bundle :cache
 
         expect(bundled_app("vendor/cache/rack-1.0.0.gem")).to exist
         expect(bundled_app("vendor/cache/rack-obama-1.0.gem")).to exist
 
-        bundle "install --deployment"
+        bundle "config --local deployment true"
+        bundle :install
 
-        expect(exitstatus).to eq(0) if exitstatus
         expect(the_bundle).to include_gems("rack-obama 1.0.0", "rack 1.0.0")
       end
     end
@@ -119,12 +122,16 @@ RSpec.describe "bundle install with gems on multiple sources" do
           build_gem "rack", "1.0.0" do |s|
             s.write "lib/rack.rb", "RACK = 'FAIL'"
           end
+
+          build_gem "rack-obama" do |s|
+            s.add_dependency "rack"
+          end
         end
 
         gemfile <<-G
-          source "file://#{gem_repo3}"
+          source "#{file_uri_for(gem_repo3)}"
           gem "rack-obama" # should come from repo3!
-          gem "rack", :source => "file://#{gem_repo1}"
+          gem "rack", :source => "#{file_uri_for(gem_repo1)}"
         G
       end
 
@@ -135,7 +142,7 @@ RSpec.describe "bundle install with gems on multiple sources" do
       end
     end
 
-    context "with an indirect dependency" do
+    context "when a pinned gem has an indirect dependency" do
       before do
         build_repo gem_repo3 do
           build_gem "depends_on_rack", "1.0.1" do |s|
@@ -152,8 +159,8 @@ RSpec.describe "bundle install with gems on multiple sources" do
           end
 
           gemfile <<-G
-            source "file://#{gem_repo2}"
-            source "file://#{gem_repo3}" do
+            source "#{file_uri_for(gem_repo2)}"
+            source "#{file_uri_for(gem_repo3)}" do
               gem "depends_on_rack"
             end
           G
@@ -181,10 +188,26 @@ RSpec.describe "bundle install with gems on multiple sources" do
             end
           end
 
-          it "installs from the same source without any warning" do
-            bundle :install
-            expect(out).not_to include("Warning")
-            expect(the_bundle).to include_gems("depends_on_rack 1.0.1", "rack 1.0.0")
+          context "when disable_multisource is set" do
+            before do
+              bundle "config set disable_multisource true"
+            end
+
+            it "installs from the same source without any warning" do
+              bundle :install
+
+              expect(out).not_to include("Warning: the gem 'rack' was found in multiple sources.")
+              expect(err).not_to include("Warning: the gem 'rack' was found in multiple sources.")
+              expect(the_bundle).to include_gems("depends_on_rack 1.0.1", "rack 1.0.0")
+
+              # when there is already a lock file, and the gems are missing, so try again
+              system_gems []
+              bundle :install
+
+              expect(out).not_to include("Warning: the gem 'rack' was found in multiple sources.")
+              expect(err).not_to include("Warning: the gem 'rack' was found in multiple sources.")
+              expect(the_bundle).to include_gems("depends_on_rack 1.0.1", "rack 1.0.0")
+            end
           end
         end
       end
@@ -200,8 +223,8 @@ RSpec.describe "bundle install with gems on multiple sources" do
         context "and not in any other sources" do
           before do
             gemfile <<-G
-              source "file://#{gem_repo2}"
-              source "file://#{gem_repo3}" do
+              source "#{file_uri_for(gem_repo2)}"
+              source "#{file_uri_for(gem_repo3)}" do
                 gem "depends_on_rack"
               end
             G
@@ -217,22 +240,26 @@ RSpec.describe "bundle install with gems on multiple sources" do
         context "and in yet another source" do
           before do
             gemfile <<-G
-              source "file://#{gem_repo1}"
-              source "file://#{gem_repo2}"
-              source "file://#{gem_repo3}" do
+              source "#{file_uri_for(gem_repo1)}"
+              source "#{file_uri_for(gem_repo2)}"
+              source "#{file_uri_for(gem_repo3)}" do
                 gem "depends_on_rack"
               end
             G
           end
 
-          it "installs from the other source and warns about ambiguous gems" do
-            bundle "config major_deprecations true"
+          it "installs from the other source and warns about ambiguous gems", :bundler => "2" do
             bundle :install
-            expect(out).to have_major_deprecation a_string_including("Your Gemfile contains multiple primary sources.")
-            expect(out).to include("Warning: the gem 'rack' was found in multiple sources.")
-            expect(out).to include("Installed from: file:#{gem_repo2}")
+            expect(err).to include("Warning: the gem 'rack' was found in multiple sources.")
+            expect(err).to include("Installed from: #{file_uri_for(gem_repo2)}")
             expect(the_bundle).to include_gems("depends_on_rack 1.0.1", "rack 1.0.0")
           end
+
+          it "fails", :bundler => "3" do
+            bundle :install, :raise_on_error => false
+            expect(err).to include("Each source after the first must include a block")
+            expect(exitstatus).to eq(4)
+          end
         end
 
         context "and only the dependency is pinned" do
@@ -245,28 +272,113 @@ RSpec.describe "bundle install with gems on multiple sources" do
             end
 
             gemfile <<-G
-              source "file://#{gem_repo3}" # contains depends_on_rack
-              source "file://#{gem_repo2}" # contains broken rack
+              source "#{file_uri_for(gem_repo3)}" # contains depends_on_rack
+              source "#{file_uri_for(gem_repo2)}" # contains broken rack
 
               gem "depends_on_rack" # installed from gem_repo3
-              gem "rack", :source => "file://#{gem_repo1}"
+              gem "rack", :source => "#{file_uri_for(gem_repo1)}"
             G
           end
 
-          it "installs the dependency from the pinned source without warning" do
+          it "installs the dependency from the pinned source without warning", :bundler => "2" do
             bundle :install
 
-            expect(out).not_to include("Warning: the gem 'rack' was found in multiple sources.")
+            expect(err).not_to include("Warning: the gem 'rack' was found in multiple sources.")
             expect(the_bundle).to include_gems("depends_on_rack 1.0.1", "rack 1.0.0")
 
-            # In https://github.com/bundler/bundler/issues/3585 this failed
+            # In https://github.com/rubygems/bundler/issues/3585 this failed
             # when there is already a lock file, and the gems are missing, so try again
             system_gems []
             bundle :install
 
-            expect(out).not_to include("Warning: the gem 'rack' was found in multiple sources.")
+            expect(err).not_to include("Warning: the gem 'rack' was found in multiple sources.")
             expect(the_bundle).to include_gems("depends_on_rack 1.0.1", "rack 1.0.0")
           end
+
+          it "fails", :bundler => "3" do
+            bundle :install, :raise_on_error => false
+            expect(err).to include("Each source after the first must include a block")
+            expect(exitstatus).to eq(4)
+          end
+        end
+      end
+    end
+
+    context "when a top-level gem has an indirect dependency" do
+      context "when disable_multisource is set" do
+        before do
+          bundle "config set disable_multisource true"
+        end
+
+        before do
+          build_repo gem_repo2 do
+            build_gem "depends_on_rack", "1.0.1" do |s|
+              s.add_dependency "rack"
+            end
+          end
+
+          build_repo gem_repo3 do
+            build_gem "unrelated_gem", "1.0.0"
+          end
+
+          gemfile <<-G
+            source "#{file_uri_for(gem_repo2)}"
+
+            gem "depends_on_rack"
+
+            source "#{file_uri_for(gem_repo3)}" do
+              gem "unrelated_gem"
+            end
+          G
+        end
+
+        context "and the dependency is only in the top-level source" do
+          before do
+            update_repo gem_repo2 do
+              build_gem "rack", "1.0.0"
+            end
+          end
+
+          it "installs all gems without warning" do
+            bundle :install
+            expect(err).not_to include("Warning")
+            expect(the_bundle).to include_gems("depends_on_rack 1.0.1", "rack 1.0.0", "unrelated_gem 1.0.0")
+          end
+        end
+
+        context "and the dependency is only in a pinned source" do
+          before do
+            update_repo gem_repo3 do
+              build_gem "rack", "1.0.0" do |s|
+                s.write "lib/rack.rb", "RACK = 'FAIL'"
+              end
+            end
+          end
+
+          it "does not find the dependency" do
+            bundle :install, :raise_on_error => false
+            expect(err).to include("Could not find gem 'rack', which is required by gem 'depends_on_rack', in any of the relevant sources")
+          end
+        end
+
+        context "and the dependency is in both the top-level and a pinned source" do
+          before do
+            update_repo gem_repo2 do
+              build_gem "rack", "1.0.0"
+            end
+
+            update_repo gem_repo3 do
+              build_gem "rack", "1.0.0" do |s|
+                s.write "lib/rack.rb", "RACK = 'FAIL'"
+              end
+            end
+          end
+
+          it "installs the dependency from the top-level source without warning" do
+            bundle :install
+            expect(err).not_to include("Warning")
+            expect(the_bundle).to include_gems("depends_on_rack 1.0.1", "rack 1.0.0", "unrelated_gem 1.0.0")
+          end
         end
       end
     end
@@ -278,25 +390,25 @@ RSpec.describe "bundle install with gems on multiple sources" do
         end
 
         gemfile <<-G
-          source "file://#{gem_repo3}"
-          gem "not_in_repo1", :source => "file://#{gem_repo1}"
+          source "#{file_uri_for(gem_repo3)}"
+          gem "not_in_repo1", :source => "#{file_uri_for(gem_repo1)}"
         G
       end
 
       it "does not install the gem" do
-        bundle :install
-        expect(out).to include("Could not find gem 'not_in_repo1'")
+        bundle :install, :raise_on_error => false
+        expect(err).to include("Could not find gem 'not_in_repo1'")
       end
     end
 
     context "with an existing lockfile" do
       before do
-        system_gems "rack-0.9.1", "rack-1.0.0"
+        system_gems "rack-0.9.1", "rack-1.0.0", :path => default_bundle_path
 
         lockfile <<-L
           GEM
-            remote: file:#{gem_repo1}
-            remote: file:#{gem_repo3}
+            remote: #{file_uri_for(gem_repo1)}
+            remote: #{file_uri_for(gem_repo3)}
             specs:
               rack (0.9.1)
 
@@ -308,14 +420,14 @@ RSpec.describe "bundle install with gems on multiple sources" do
         L
 
         gemfile <<-G
-          source "file://#{gem_repo1}"
-          source "file://#{gem_repo3}" do
+          source "#{file_uri_for(gem_repo1)}"
+          source "#{file_uri_for(gem_repo3)}" do
             gem 'rack'
           end
         G
       end
 
-      # Reproduction of https://github.com/bundler/bundler/issues/3298
+      # Reproduction of https://github.com/rubygems/bundler/issues/3298
       it "does not unlock the installed gem on exec" do
         expect(the_bundle).to include_gems("rack 0.9.1")
       end
@@ -326,7 +438,7 @@ RSpec.describe "bundle install with gems on multiple sources" do
         build_lib "foo"
 
         gemfile <<-G
-          gem "rack", :source => "file://#{gem_repo1}"
+          gem "rack", :source => "#{file_uri_for(gem_repo1)}"
           gem "foo", :path => "#{lib_path("foo-1.0")}"
         G
       end
@@ -337,7 +449,6 @@ RSpec.describe "bundle install with gems on multiple sources" do
         bundle %(exec ruby -e 'puts "OK"')
 
         expect(out).to include("OK")
-        expect(exitstatus).to eq(0) if exitstatus
       end
     end
   end
@@ -347,56 +458,38 @@ RSpec.describe "bundle install with gems on multiple sources" do
       system_gems "rack-0.9.1"
 
       gemfile <<-G
-        source "file://#{gem_repo1}"
+        source "#{file_uri_for(gem_repo1)}"
         gem "rack" # shoud come from repo1!
       G
     end
 
     it "installs the gems without any warning" do
       bundle :install
-      expect(out).not_to include("Warning")
+      expect(err).not_to include("Warning")
       expect(the_bundle).to include_gems("rack 1.0.0")
     end
   end
 
   context "when a single source contains multiple locked gems" do
     before do
-      # 1. With these gems,
+      # With these gems,
       build_repo4 do
         build_gem "foo", "0.1"
         build_gem "bar", "0.1"
       end
 
-      # 2. Installing this gemfile will produce...
+      # Installing this gemfile...
       gemfile <<-G
-        source 'file://#{gem_repo1}'
+        source '#{file_uri_for(gem_repo1)}'
         gem 'rack'
-        gem 'foo', '~> 0.1', :source => 'file://#{gem_repo4}'
-        gem 'bar', '~> 0.1', :source => 'file://#{gem_repo4}'
+        gem 'foo', '~> 0.1', :source => '#{file_uri_for(gem_repo4)}'
+        gem 'bar', '~> 0.1', :source => '#{file_uri_for(gem_repo4)}'
       G
 
-      # 3. this lockfile.
-      lockfile <<-L
-        GEM
-          remote: file:/Users/andre/src/bundler/bundler/tmp/gems/remote1/
-          remote: file:/Users/andre/src/bundler/bundler/tmp/gems/remote4/
-          specs:
-            bar (0.1)
-            foo (0.1)
-            rack (1.0.0)
+      bundle "config --local path ../gems/system"
+      bundle :install
 
-        PLATFORMS
-          ruby
-
-        DEPENDENCIES
-          bar (~> 0.1)!
-          foo (~> 0.1)!
-          rack
-      L
-
-      bundle "install --path ../gems/system"
-
-      # 4. Then we add some new versions...
+      # And then we add some new versions...
       update_repo4 do
         build_gem "foo", "0.2"
         build_gem "bar", "0.3"
@@ -404,17 +497,16 @@ RSpec.describe "bundle install with gems on multiple sources" do
     end
 
     it "allows them to be unlocked separately" do
-      # 5. and install this gemfile, updating only foo.
+      # And install this gemfile, updating only foo.
       install_gemfile <<-G
-        source 'file://#{gem_repo1}'
+        source '#{file_uri_for(gem_repo1)}'
         gem 'rack'
-        gem 'foo', '~> 0.2', :source => 'file://#{gem_repo4}'
-        gem 'bar', '~> 0.1', :source => 'file://#{gem_repo4}'
+        gem 'foo', '~> 0.2', :source => '#{file_uri_for(gem_repo4)}'
+        gem 'bar', '~> 0.1', :source => '#{file_uri_for(gem_repo4)}'
       G
 
-      # 6. Which should update foo to 0.2, but not the (locked) bar 0.1
-      expect(the_bundle).to include_gems("foo 0.2")
-      expect(the_bundle).to include_gems("bar 0.1")
+      # It should update foo to 0.2, but not the (locked) bar 0.1
+      expect(the_bundle).to include_gems("foo 0.2", "bar 0.1")
     end
   end
 
@@ -428,10 +520,10 @@ RSpec.describe "bundle install with gems on multiple sources" do
         build_git "git2"
 
         install_gemfile <<-G
-          source "file://#{gem_repo1}"
+          source "#{file_uri_for(gem_repo1)}"
           gem "rails"
 
-          source "file://#{gem_repo3}" do
+          source "#{file_uri_for(gem_repo3)}" do
             gem "rack"
           end
 
@@ -452,16 +544,16 @@ RSpec.describe "bundle install with gems on multiple sources" do
 
   context "when a gem is installed to system gems" do
     before do
-      install_gemfile! <<-G
-        source "file://#{gem_repo1}"
+      install_gemfile <<-G
+        source "#{file_uri_for(gem_repo1)}"
         gem "rack"
       G
     end
 
     context "and the gemfile changes" do
       it "is still able to find that gem from remote sources" do
-        source_uri = "file://#{gem_repo1}"
-        second_uri = "file://#{gem_repo4}"
+        source_uri = file_uri_for(gem_repo1)
+        second_uri = file_uri_for(gem_repo4)
 
         build_repo4 do
           build_gem "rack", "2.0.1.1.forked"
@@ -469,7 +561,7 @@ RSpec.describe "bundle install with gems on multiple sources" do
         end
 
         # When this gemfile is installed...
-        gemfile <<-G
+        install_gemfile <<-G
           source "#{source_uri}"
 
           source "#{second_uri}" do
@@ -479,26 +571,6 @@ RSpec.describe "bundle install with gems on multiple sources" do
           gem "rack-obama"
         G
 
-        # It creates this lockfile.
-        lockfile <<-L
-          GEM
-            remote: #{source_uri}/
-            remote: #{second_uri}/
-            specs:
-              rack (2.0.1.1.forked)
-              rack-obama (1.0)
-                rack
-              thor (0.19.1.1.forked)
-
-          PLATFORMS
-            ruby
-
-          DEPENDENCIES
-            rack (= 2.0.1.1.forked)!
-            rack-obama
-            thor!
-        L
-
         # Then we change the Gemfile by adding a version to thor
         gemfile <<-G
           source "#{source_uri}"
@@ -511,8 +583,67 @@ RSpec.describe "bundle install with gems on multiple sources" do
         G
 
         # But we should still be able to find rack 2.0.1.1.forked and install it
-        bundle! :install
+        bundle :install
       end
     end
   end
+
+  describe "source changed to one containing a higher version of a dependency" do
+    before do
+      install_gemfile <<-G
+        source "#{file_uri_for(gem_repo1)}"
+
+        gem "rack"
+      G
+
+      build_repo2 do
+        build_gem "bar"
+      end
+
+      build_lib("gemspec_test", :path => tmp.join("gemspec_test")) do |s|
+        s.add_dependency "bar", "=1.0.0"
+      end
+
+      install_gemfile <<-G
+        source "#{file_uri_for(gem_repo2)}"
+        gem "rack"
+        gemspec :path => "#{tmp.join("gemspec_test")}"
+      G
+    end
+
+    it "keeps the old version", :bundler => "2" do
+      expect(the_bundle).to include_gems("rack 1.0.0")
+    end
+
+    it "installs the higher version in the new repo", :bundler => "3" do
+      expect(the_bundle).to include_gems("rack 1.2")
+    end
+  end
+
+  context "when a gem is available from multiple ambiguous sources", :bundler => "3" do
+    it "raises, suggesting a source block" do
+      build_repo4 do
+        build_gem "depends_on_rack" do |s|
+          s.add_dependency "rack"
+        end
+        build_gem "rack"
+      end
+
+      install_gemfile <<-G, :raise_on_error => false
+        source "#{file_uri_for(gem_repo4)}"
+        source "#{file_uri_for(gem_repo1)}" do
+          gem "thin"
+        end
+        gem "depends_on_rack"
+      G
+      expect(last_command).to be_failure
+      expect(err).to eq strip_whitespace(<<-EOS).strip
+        The gem 'rack' was found in multiple relevant sources.
+          * rubygems repository #{file_uri_for(gem_repo1)}/ or installed locally
+          * rubygems repository #{file_uri_for(gem_repo4)}/ or installed locally
+        You must add this gem to the source block for the source you wish it to be installed from.
+      EOS
+      expect(the_bundle).not_to be_locked
+    end
+  end
 end
